@@ -14,9 +14,6 @@ This extension:
 - enable the changeset-evolution feature for Mercurial,
 - improves some aspect of the early implementation in Mercurial core,
 
-Note that a version dedicated to server usage only (no local working copy) is
-available as 'evolve.serveronly'.
-
 While many feature related to changeset evolution are directly handled by core
 this extensions contains significant additions recommended to any user of
 changeset evolution.
@@ -33,7 +30,7 @@ Some improvement and bug fixes available in newer version of Mercurial are also
 backported to older version of Mercurial by this extension. Some older
 experimental protocols are also supported for a longer time in the extension to
 help people transitioning. (The extension is currently compatible down to
-Mercurial version 4.8).
+Mercurial version 4.9).
 
 New Config::
 
@@ -490,9 +487,14 @@ def setupparentcommand(ui):
 
 @eh.uisetup
 def _installalias(ui):
+    odiffalias = b"diff --hidden --rev 'limit(predecessors(.),1)' --rev ."
     if ui.config(b'alias', b'odiff', None) is None:
         ui.setconfig(b'alias', b'odiff',
-                     b"diff --hidden --rev 'limit(predecessors(.),1)' --rev .",
+                     odiffalias,
+                     b'evolve')
+    if ui.config(b'alias', b'obsdiff', None) is None:
+        ui.setconfig(b'alias', b'obsdiff',
+                     odiffalias,
                      b'evolve')
 
 #####################################################################
