@@ -1417,9 +1417,10 @@ def cmdpick(ui, repo, *revs, **opts):
 def _dopick(ui, repo, pickstate, origctx):
     """shared logic for performing or continuing a pick"""
     overrides = {(b'phases', b'new-commit'): origctx.phase()}
+    new_desc = evolvecmd._rewrite_commit_message_hashes(repo,
+                                                        origctx.description())
     with repo.ui.configoverride(overrides, b'pick'):
-        newnode = repo.commit(text=origctx.description(),
-                              user=origctx.user(),
+        newnode = repo.commit(text=new_desc, user=origctx.user(),
                               date=origctx.date(), extra=origctx.extra())
     compat.setbranch(repo, origctx.branch())
 
