@@ -20,7 +20,6 @@ from mercurial import (
 from mercurial.utils.stringutil import forcebytestr
 
 from . import (
-    compat,
     error,
     exthelper,
     genericcaches,
@@ -86,17 +85,15 @@ class firstmergecache(genericcaches.changelogsourcebase):
         cl = repo.unfiltered().changelog
         total = len(data)
 
-        def progress(pos, rev=None):
-            revstr = b'' if rev is None else (b'rev %d' % rev)
-            compat.progress(repo.ui, b'updating firstmerge cache',
-                            pos, revstr, unit=_(b'changesets'), total=total)
-        progress(0)
+        progress = repo.ui.makeprogress(b'updating firstmerge cache', _(b'changesets'), total)
+        progress.update(0)
         for idx, rev in enumerate(data, 1):
             assert rev == len(self._data), (rev, len(self._data))
             self._data.append(self._firstmerge(cl, rev))
             if not (idx % 10000): # progress as a too high performance impact
-                progress(idx, rev)
-        progress(None)
+                revstr = b'' if rev is None else (b'rev %d' % rev)
+                progress.update(idx, item=revstr)
+        progress.complete()
 
     def _firstmerge(self, changelog, rev):
         cl = changelog
