@@ -11,6 +11,8 @@
 #   happy one piece of it (and hopefully, able to reuse it in other core
 #   commands).
 
+import functools
+
 from mercurial import (
     cmdutil,
     commands,
@@ -60,8 +62,13 @@ def precheck(repo, revs, action=b'rewrite', check_divergence=True):
 
     <action> can be used to control the commit message.
     """
+    if not pycompat.ispy3 and isinstance(corerewriteutil.precheck, functools.partial):
+        # inspect.getargspec() on py2 cannot inspect functools.partial objects
+        # directly, so we need to provide it with the underlying function
+        args = pycompat.getargspec(corerewriteutil.precheck.func).args
+    else:
+        args = pycompat.getargspec(corerewriteutil.precheck).args
     # hg <= 6.1 (d4752aeb20f1)
-    args = pycompat.getargspec(corerewriteutil.precheck).args
     if r'check_divergence' in args:
         return corerewriteutil.precheck(repo, revs, action,
                                         check_divergence=check_divergence)
