@@ -165,6 +165,7 @@ import weakref
 from mercurial.i18n import _
 from mercurial import (
     bookmarks,
+    bundlerepo,
     changelog,
     cmdutil,
     commands,
@@ -186,7 +187,9 @@ from mercurial import (
     phases,
     pycompat,
     registrar,
+    rewriteutil,
     scmutil,
+    smartset,
     templatefilters,
     util,
 )
@@ -233,10 +236,10 @@ colortable = {b'topic.active': b'green',
               b'log.topic': b'green_background',
               }
 
-__version__ = b'1.0.1'
+__version__ = b'1.1.0'
 
-testedwith = b'4.8 4.9 5.0 5.1 5.2 5.3 5.4 5.5 5.6 5.7 5.8 5.9 6.0 6.1 6.2 6.3 6.4'
-minimumhgversion = b'4.8'
+testedwith = b'4.9 5.0 5.1 5.2 5.3 5.4 5.5 5.6 5.7 5.8 5.9 6.0 6.1 6.2 6.3 6.4 6.5'
+minimumhgversion = b'4.9'
 buglink = b'https://bz.mercurial-scm.org/'
 
 configtable = {}
@@ -269,6 +272,22 @@ configitem(b'_internal', b'tns-disable-fqbn',
 configitem(b'_internal', b'tns-publish',
            default=False,
 )
+# used for signaling that the current command has explicit target arguments
+# (e.g. --rev or --branch) and we should ignore tns-default-* config
+configitem(b'_internal', b'tns-explicit-target',
+           default=False,
+)
+configitem(b'devel', b'tns-report-transactions',
+           default=lambda: [],
+)
+# used for allowing users to rewrite history only in their "own" topic
+# namespaces
+configitem(b'experimental', b'tns-allow-rewrite',
+           default=configitems.dynamicdefault,
+)
+configitem(b'experimental', b'tns-default-pull-namespaces',
+           default=configitems.dynamicdefault,
+)
 configitem(b'experimental', b'topic-mode.server',
            default=configitems.dynamicdefault,
 )
@@ -291,24 +310,27 @@ def extsetup(ui):
                   default=None,
         )
     if (b'devel' not in ui._knownconfig
-            or not ui._knownconfig[b'devel'].get(b'random')):
+            or not ui._knownconfig[b'devel'].get(b'randomseed')):
         extraitem(b'devel', b'randomseed',
                   default=None,
         )
 
 def _contexttns(self, force=False):
     if not force and not self.mutable():
-        return b'default'
+        return b'none'
     cache = getattr(self._repo, '_tnscache', None)
     # topic loaded, but not enabled (eg: multiple repo in the same process)
     if cache is None:
-        return b'default'
+        return b'none'
+    # topic namespace is meaningless when topic is not set
+    if not self.topic(force):
+        return b'none'
     if self.rev() is None:
         # don't cache volatile ctx instances that aren't stored on-disk yet
-        return self.extra().get(b'topic-namespace', b'default')
+        return self.extra().get(b'topic-namespace', b'none')
     tns = cache.get(self.rev())
     if tns is None:
-        tns = self.extra().get(b'topic-namespace', b'default')
+        tns = self.extra().get(b'topic-namespace', b'none')
         self._repo._tnscache[self.rev()] = tns
     return tns
 
@@ -430,6 +452,70 @@ def wrapwctxdirty(orig, self, missing=False, merge=True, branch=True):
         or (missing and self.deleted())
     )
 
+def find_affected_tns(repo, tr):
+    origrepolen = tr.changes[b'origrepolen']
+    unfi = repo.unfiltered()
+
+    affected = set()
+    # These are the new changesets that weren't in the repo before this
+    # transaction
+    for rev in smartset.spanset(repo, start=origrepolen):
+        ctx = unfi[rev]
+        tns = ctx.topic_namespace()
+        affected.add(tns)
+
+    # These are the changesets obsoleted by this transaction
+    for rev in obsutil.getobsoleted(repo, tr):
+        ctx = unfi[rev]
+        tns = ctx.topic_namespace()
+        affected.add(tns)
+
+    # Phase movements, we only care about:
+    # - publishing changesets (since they lose topic namespace)
+    # - forcefully making changesets draft again
+    # - turning secret changesets draft and making them visible to peers
+    tnsphases = (phases.secret, phases.draft)
+    phasechanges = tr.changes[b'phases']
+    if isinstance(phasechanges, dict):
+        # hg <= 5.3 (fdc802f29b2c)
+        phasechanges = [((k,), v) for k, v in phasechanges.items()]
+    for revs, (old, new) in phasechanges:
+        if old not in tnsphases and new not in tnsphases:
+            # Skip phase movement if there is no phase (old or new) that has
+            # visible topic namespace (i.e. draft and secret)
+            continue
+        revs = [rev for rev in revs if rev < origrepolen]
+        for rev in revs:
+            ctx = unfi[rev]
+            tns = ctx.topic_namespace(force=True)
+            affected.add(tns)
+
+    # We want to detect any bookmark movement, even within one topic namespace
+    for name, nodes in tr.changes[b'bookmarks'].items():
+        for n in nodes:
+            if n is not None and n in unfi:
+                ctx = unfi[n]
+                tns = ctx.topic_namespace()
+                affected.add(tns)
+
+    # We don't care about changesets without topic namespace
+    affected.discard(b'none')
+
+    tr.changes[b'tns'] = affected
+    report_affected_tns(repo, tr)
+
+def report_affected_tns(repo, tr):
+    report = set(repo.ui.configlist(b'devel', b'tns-report-transactions'))
+    # transaction names sometimes also have a URL after a newline byte
+    trnames = (trname.partition(b'\n')[0] for trname in tr._names)
+    if b'*' not in report:
+        # * matches any transaction
+        if not any(trname in report for trname in trnames):
+            return
+
+    if tr.changes[b'tns']:
+        repo.ui.status(b'topic namespaces affected: %s\n' % b' '.join(sorted(tr.changes[b'tns'])))
+
 def uisetup(ui):
     destination.modsetup(ui)
     discovery.modsetup(ui)
@@ -506,8 +592,37 @@ def uisetup(ui):
         extensions.wrapfunction(histedit.histeditaction, 'applychange',
                                 applychangewrap)
 
+    # Wrapping precheck() both in core and in evolve to make sure all rewrite
+    # operations that could use precheck() are covered
+    extensions.wrapfunction(rewriteutil, 'precheck', wrapprecheck)
+    try:
+        evolve = extensions.find(b'evolve')
+        extensions.wrapfunction(evolve.rewriteutil, 'precheck', wrapprecheck)
+    except (KeyError, AttributeError):
+        pass
+
     server.setupserver(ui)
 
+    # We want bundle repos to also have caches for topic extension, because we
+    # want to, for example, see topic and topic namespaces in `hg incoming`
+    # regardless if the bundle repo has topic extension, as long as local repo
+    # has topic enabled.
+    class topicbundlerepo(bundlerepo.bundlerepository):
+        @util.propertycache
+        def _tnscache(self):
+            return {}
+
+        @util.propertycache
+        def _topiccache(self):
+            return {}
+
+        def invalidatecaches(self):
+            self._tnscache.clear()
+            self._topiccache.clear()
+            super(topicbundlerepo, self).invalidatecaches()
+
+    bundlerepo.bundlerepository = topicbundlerepo
+
 def reposetup(ui, repo):
     if not isinstance(repo, localrepo.localrepository):
         return # this can be a peer in the ssh case (puzzling)
@@ -541,6 +656,7 @@ def reposetup(ui, repo):
             else:
                 mode = b'none'
             caps.add(b'ext-topics-publish=%s' % mode)
+            caps.add(b'ext-topics-tns-heads')
             return caps
 
         def commit(self, *args, **kwargs):
@@ -556,7 +672,7 @@ def reposetup(ui, repo):
         def commitctx(self, ctx, *args, **kwargs):
             if isinstance(ctx, context.workingcommitctx):
                 current = self.currenttopic
-                if current:
+                if current and constants.extrakey not in ctx.extra():
                     ctx.extra()[constants.extrakey] = current
             return super(topicrepo, self).commitctx(ctx, *args, **kwargs)
 
@@ -568,15 +684,16 @@ def reposetup(ui, repo):
         def topic_namespaces(self):
             if self._topic_namespaces is not None:
                 return self._topic_namespaces
-            namespaces = set([self.currenttns])
+            namespaces = set([b'none', self.currenttns])
             for c in self.set(b'not public()'):
                 namespaces.add(c.topic_namespace())
+            namespaces.remove(b'none')
             self._topic_namespaces = namespaces
             return namespaces
 
         @property
         def currenttns(self):
-            return self.vfs.tryread(b'topic-namespace') or b'default'
+            return self.vfs.tryread(b'topic-namespace') or b'none'
 
         @util.propertycache
         def _topiccache(self):
@@ -636,6 +753,7 @@ def reposetup(ui, repo):
                                                       closed=closed)
 
         def invalidatecaches(self):
+            self._tnscache.clear()
             self._topiccache.clear()
             super(topicrepo, self).invalidatecaches()
 
@@ -656,6 +774,20 @@ def reposetup(ui, repo):
                     def branchmaptns(self):
                         usetopic = not self._repo.publishing()
                         return self._repo.branchmaptns(topic=usetopic)
+
+                    def tns_heads(self, namespaces):
+                        if b'*' in namespaces:
+                            # pulling all topic namespaces, all changesets are visible
+                            return self._repo.heads()
+                        else:
+                            # only changesets in the selected topic namespaces are visible
+                            h = []
+                            entries = compat.bcentries(self._repo.branchmaptns())
+                            for branch, nodes in compat.branchmapitems(entries):
+                                namedbranch, tns, topic = common.parsefqbn(branch)
+                                if tns == b'none' or tns in namespaces:
+                                    h.extend(nodes)
+                            return h
                 peer.__class__ = topicpeer
             return peer
 
@@ -669,21 +801,21 @@ def reposetup(ui, repo):
             if self.ui.configbool(b'experimental', b'enforce-single-head'):
                 if util.safehasattr(tr, '_validator'):
                     # hg <= 5.3 (36f08ae87ef6)
-                    origvalidator = tr._validator
+                    origvalidator_single_head = tr._validator
 
-                def _validate(tr2):
+                def _validate_single_head(tr2):
                     repo = reporef()
                     flow.enforcesinglehead(repo, tr2)
 
                 def validator(tr2):
-                    _validate(tr2)
-                    return origvalidator(tr2)
+                    _validate_single_head(tr2)
+                    return origvalidator_single_head(tr2)
 
                 if util.safehasattr(tr, '_validator'):
                     # hg <= 5.3 (36f08ae87ef6)
                     tr._validator = validator
                 else:
-                    tr.addvalidator(b'000-enforce-single-head', _validate)
+                    tr.addvalidator(b'000-enforce-single-head', _validate_single_head)
 
             topicmodeserver = self.ui.config(b'experimental',
                                              b'topic-mode.server', b'ignore')
@@ -693,21 +825,21 @@ def reposetup(ui, repo):
             if (topicmodeserver != b'ignore' and ispush):
                 if util.safehasattr(tr, '_validator'):
                     # hg <= 5.3 (36f08ae87ef6)
-                    origvalidator = tr._validator
+                    origvalidator_untopiced = tr._validator
 
-                def _validate(tr2):
+                def _validate_untopiced(tr2):
                     repo = reporef()
                     flow.rejectuntopicedchangeset(repo, tr2)
 
                 def validator(tr2):
-                    _validate(tr2)
-                    return origvalidator(tr2)
+                    _validate_untopiced(tr2)
+                    return origvalidator_untopiced(tr2)
 
                 if util.safehasattr(tr, '_validator'):
                     # hg <= 5.3 (36f08ae87ef6)
                     tr._validator = validator
                 else:
-                    tr.addvalidator(b'000-reject-untopiced', _validate)
+                    tr.addvalidator(b'000-reject-untopiced', _validate_untopiced)
 
             elif publishbare and ispush:
                 origclose = tr.close
@@ -725,21 +857,40 @@ def reposetup(ui, repo):
             if not allow_publish:
                 if util.safehasattr(tr, '_validator'):
                     # hg <= 5.3 (36f08ae87ef6)
-                    origvalidator = tr._validator
+                    origvalidator_publish = tr._validator
 
-                def _validate(tr2):
+                def _validate_publish(tr2):
                     repo = reporef()
                     flow.reject_publish(repo, tr2)
 
                 def validator(tr2):
-                    _validate(tr2)
-                    return origvalidator(tr2)
+                    _validate_publish(tr2)
+                    return origvalidator_publish(tr2)
 
                 if util.safehasattr(tr, '_validator'):
                     # hg <= 5.3 (36f08ae87ef6)
                     tr._validator = validator
                 else:
-                    tr.addvalidator(b'000-reject-publish', _validate)
+                    tr.addvalidator(b'000-reject-publish', _validate_publish)
+
+            if util.safehasattr(tr, '_validator'):
+                # hg <= 5.3 (36f08ae87ef6)
+                origvalidator_affected_tns = tr._validator
+
+            def _validate_affected_tns(tr2):
+                repo = reporef()
+                find_affected_tns(repo, tr2)
+
+            def validator(tr2):
+                result = origvalidator_affected_tns(tr2)
+                _validate_affected_tns(tr2)
+                return result
+
+            if util.safehasattr(tr, '_validator'):
+                # hg <= 5.3 (36f08ae87ef6)
+                tr._validator = validator
+            else:
+                tr.addvalidator(b'999-find-affected-tns', _validate_affected_tns)
 
             # real transaction start
             ct = self.currenttopic
@@ -817,7 +968,7 @@ def wrapinit(orig, self, repo, *args, **kwargs):
             self._extra[b'topic-namespace'] = repo.currenttns
         else:
             # Default value will be dropped from extra by another hack at the changegroup level
-            self._extra[b'topic-namespace'] = b'default'
+            self._extra[b'topic-namespace'] = b'none'
     if constants.extrakey not in self._extra:
         if getattr(repo, 'currenttopic', b''):
             self._extra[constants.extrakey] = repo.currenttopic
@@ -828,12 +979,16 @@ def wrapinit(orig, self, repo, *args, **kwargs):
 def wrapadd(orig, cl, manifest, files, desc, transaction, p1, p2, user,
             date=None, extra=None, p1copies=None, p2copies=None,
             filesadded=None, filesremoved=None):
-    if b'topic-namespace' in extra and extra[b'topic-namespace'] == b'default':
+    if b'topic-namespace' in extra and extra[b'topic-namespace'] == b'none':
         extra = extra.copy()
         del extra[b'topic-namespace']
     if constants.extrakey in extra and not extra[constants.extrakey]:
         extra = extra.copy()
         del extra[constants.extrakey]
+    if constants.extrakey not in extra and b'topic-namespace' in extra:
+        # if topic is not in extra, drop namespace as well
+        extra = extra.copy()
+        del extra[b'topic-namespace']
     # hg <= 4.9 (0e41f40b01cc)
     kwargs = {}
     if p1copies is not None:
@@ -1518,14 +1673,14 @@ def mergeupdatewrap(orig, repo, node, branchmerge, force, *args, **kwargs):
         if repo.ui.configbool(b'_internal', b'keep-topic'):
             ist0 = True
         if ((not partial and not branchmerge) or isrebase) and not ist0:
-            tns = b'default'
+            tns = b'none'
             t = b''
             pctx = repo[node]
             if pctx.phase() > phases.public:
                 tns = pctx.topic_namespace()
                 t = pctx.topic()
             repo.vfs.write(b'topic-namespace', tns)
-            if tns != b'default' and tns != otns:
+            if tns != b'none' and tns != otns:
                 repo.ui.status(_(b"switching to topic-namespace %s\n") % tns)
             repo.vfs.write(b'topic', t)
             if t and t != ot:
@@ -1584,22 +1739,37 @@ def _fixrebase(loaded):
 
 ## preserve topic during import/export
 
+def _exporttns(seq, ctx):
+    tns = ctx.topic_namespace()
+    if tns != b'none':
+        return b'EXP-Topic-Namespace %s' % tns
+    return None
+
 def _exporttopic(seq, ctx):
     topic = ctx.topic()
     if topic:
         return b'EXP-Topic %s' % topic
     return None
 
+def _importtns(repo, patchdata, extra, opts):
+    if b'topic-namespace' in patchdata:
+        extra[b'topic-namespace'] = patchdata[b'topic-namespace']
+
 def _importtopic(repo, patchdata, extra, opts):
     if b'topic' in patchdata:
         extra[b'topic'] = patchdata[b'topic']
 
 def setupimportexport(ui):
     """run at ui setup time to install import/export logic"""
+    cmdutil.extraexport.append(b'topic-namespace')
+    cmdutil.extraexportmap[b'topic-namespace'] = _exporttns
     cmdutil.extraexport.append(b'topic')
     cmdutil.extraexportmap[b'topic'] = _exporttopic
+    cmdutil.extrapreimport.append(b'topic-namespace')
+    cmdutil.extrapreimportmap[b'topic-namespace'] = _importtns
     cmdutil.extrapreimport.append(b'topic')
     cmdutil.extrapreimportmap[b'topic'] = _importtopic
+    patch.patchheadermap.append((b'EXP-Topic-Namespace', b'topic-namespace'))
     patch.patchheadermap.append((b'EXP-Topic', b'topic'))
 
 ## preserve topic during split
@@ -1617,13 +1787,33 @@ def wrappresplitupdate(original, repo, ui, prev, ctx):
     original(repo, ui, prev, ctx)
 
     # Restore the topic if need
-    if tns:
+    if tns != b'none':
         _changecurrenttns(repo, tns)
     if topic:
         _changecurrenttopic(repo, topic)
 
+def wrapprecheck(orig, repo, revs, action=b'rewrite', check_divergence=True):
+    # hg <= 6.1 (d4752aeb20f1)
+    args = pycompat.getargspec(orig).args
+    if r'check_divergence' in args:
+        orig(repo, revs, action, check_divergence=check_divergence)
+    else:
+        orig(repo, revs, action)
+
+    # TODO: at some point in future the default will change from '*' to the
+    # default topic namespace for the current user
+    allow = set(repo.ui.configlist(b'experimental', b'tns-allow-rewrite', [b'*']))
+    if b'*' not in allow:
+        namespaces = set(repo[rev].topic_namespace() for rev in revs)
+        disallowed = namespaces - allow
+        if disallowed:
+            msg = _(b"refusing to %s changesets with these topic namespaces: %s")
+            msg %= (action, b' '.join(disallowed))
+            hint = _(b"modify experimental.tns-allow-rewrite to allow rewriting changesets from these topic namespaces")
+            raise compat.InputError(msg, hint=hint)
+
 def _changecurrenttns(repo, tns):
-    if tns:
+    if tns != b'none':
         with repo.wlock():
             repo.vfs.write(b'topic-namespace', tns)
     else:
@@ -1638,7 +1828,7 @@ def debugtopicnamespace(ui, repo, tns=None, **opts):
     if opts.get('clear'):
         if tns:
             raise error.Abort(_(b"cannot use --clear when setting a topic namespace"))
-        tns = None
+        tns = b'none'
     elif not tns:
         ui.write(b'%s\n' % repo.currenttns)
         return
@@ -1651,7 +1841,7 @@ def debugtopicnamespace(ui, repo, tns=None, **opts):
         scmutil.checknewlabel(repo, tns, b'topic namespace')
     ctns = repo.currenttns
     _changecurrenttns(repo, tns)
-    if ctns == b'default' and tns:
+    if ctns == b'none' and tns != b'none':
         repo.ui.status(_(b'marked working directory as topic namespace: %s\n')
                        % tns)
 
